#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "cache_sim.h"

cache_set* build_cache(int sets, int assoc, int block_size);
void print_cache(cache_set *cache, int sets, int assoc, int block_size);
unsigned char get_byte(cache_set *cache, int set, int line, int offset);
unsigned int get_valid(cache_set *cache, int set, int line);
unsigned int get_tag(cache_set *cache, int set, int line);
void read_cache(cache_set *cache, int sets, int assoc, int block_size, unsigned int addr);
double log_2(unsigned int n);
unsigned int get_offset_val(unsigned int addr, unsigned int offset_bits);
unsigned int get_index_val(unsigned int addr, unsigned int offset_bits, unsigned int index_bits);
unsigned int get_tag_val(unsigned int addr, unsigned int offset_bits, unsigned int index_bits);
void free_cache(cache_set *cache, int sets, int assoc, int block_size);

int main(int argc, char **argv) {
    
    if (argc != 2) {
        fprintf(stderr, "Usage: %s <cache_descriptor_file>\n", argv[0]);
        exit(1);
    }

    FILE *fp = fopen(argv[1], "r");
    if (!fp) {
        fprintf(stderr, "Couldn't open %s\n", argv[1]);
        exit(1);
    }

    int sets, assoc, block_size;
    if (fscanf(fp, "%d %d %d", &sets, &assoc, &block_size) != 3) {
        fprintf(stderr, "Couldn't parse %s\n", argv[1]);
        exit(1);
    }

    cache_set *cache = build_cache(sets, assoc, block_size);
    
    int set = 0;
    int line = 0;
    int offset = 0;
    unsigned char *block = (unsigned char *)malloc(block_size);
    fscanf(fp, "%u", &cache[set].lines[line].valid);
    fscanf(fp, "%X", &cache[set].lines[line].tag);
    while (fscanf(fp, "%X", &cache[set].lines[line].block[offset]) == 1) {
        if (++offset == block_size) {
            offset = 0;
            if (++line == assoc) {
                line = 0;
                if (++set == sets) {
                    break;
                } else {
                    fscanf(fp, "%u", &cache[set].lines[line].valid);
                    fscanf(fp, "%X", &cache[set].lines[line].tag);
                }
            } else {
                fscanf(fp, "%u", &cache[set].lines[line].valid);
                fscanf(fp, "%X", &cache[set].lines[line].tag);
            }
        }
    }

    unsigned int addr;
    print_cache(cache, sets, assoc, block_size);
    while (scanf("%X", &addr) == 1) {
        read_cache(cache, sets, assoc, block_size, addr);
    }
    
    free_cache(cache, sets, assoc, block_size);

    return 0;
}

cache_set* build_cache(int sets, int assoc, int block_size) {
    cache_set *cache = (cache_set *)malloc(sets * sizeof(cache_set));
    if (!cache) {
        fprintf(stderr, "Ran out of heap space\n");
        exit(1);
    }
    for (int i = 0; i < sets; ++i) {
        cache[i].lines = (cache_line*)malloc(assoc * sizeof(cache_line));
        if (!cache[i].lines) {
            fprintf(stderr, "Ran out of heap space\n");
            exit(1);
        } else {
            for (int j = 0; j < assoc; ++j) {
                cache[i].lines[j].block = (unsigned char *)malloc(block_size);
                if (!cache[i].lines[j].block) {
                    fprintf(stderr, "Ran out of heap space\n");
                    exit(1);
                }
            }
        }
    }
    return cache;
}

void free_cache(cache_set *cache, int sets, int assoc, int block_size) {
    for (int i = 0; i < sets; ++i) {
        for (int j = 0; j < assoc; ++j) {
            free(cache[i].lines[j].block);
        }
        free(cache[i].lines);
    }
    free(cache);
}

void print_cache(cache_set *cache, int sets, int assoc, int block_size) {
    for (int i = 0; i < sets; ++i) {
        printf("Set index: %u\n", i);
        for (int j = 0; j < assoc; ++j) {
            printf("Line #%u: ", j);
            printf("valid: %u, tag: 0x%X\n", get_valid(cache, i, j), get_tag(cache, i, j));
            printf("Block: [");
            for (int k = 0; k < block_size; ++k) {
                printf("0x%X", get_byte(cache, i, j, k));
                if (k != block_size - 1)
                    printf(", ");
            }
            printf("]\n");
        }
    }
}

unsigned char get_byte(cache_set *cache, int set, int line, int offset) {
    return cache[set].lines[line].block[offset];
}

unsigned int get_valid(cache_set *cache, int set, int line) {
    return cache[set].lines[line].valid;   
}

unsigned int get_tag(cache_set *cache, int set, int line) {
    return cache[set].lines[line].tag;
}

void read_cache(cache_set *cache, int sets, int assoc, int block_size, unsigned int addr) {
    unsigned int offset_bits = (int)log_2(block_size);
    unsigned int index_bits = (int)log_2(sets);
    unsigned int offset = get_offset_val(addr, offset_bits);
    unsigned int index = get_index_val(addr, offset_bits, index_bits);
    unsigned int tag = get_tag_val(addr, offset_bits, index_bits);
    for (int i = 0; i < assoc; ++i) {
        if (get_valid(cache, index, i)) {
            if (get_tag(cache, index, i) == tag) {
                printf("Address 0x%X is a hit! Returns 0x%X\n", addr, get_byte(cache, index, i, offset));
                return;
            }
        }
    }
    printf("Address 0x%X is a miss!\n", addr);
}

double log_2(unsigned int n) {
    return log((double)n) / log(2.0);
}

unsigned int get_offset_val(unsigned int addr, unsigned int offset_bits) {
    unsigned int mask = 0;
    for (int i = 0; i < offset_bits; ++i)
        mask |= 0x1 << i;
    return addr & mask;
}

unsigned int get_index_val(unsigned int addr, unsigned int offset_bits, unsigned int index_bits) {
    unsigned int mask = 0;
    for (int i = 0; i < index_bits; ++i)
        mask |= 0x1 << i;
    mask = mask << offset_bits;
    return (addr & mask) >> offset_bits;
}

unsigned int get_tag_val(unsigned int addr, unsigned int offset_bits, unsigned int index_bits) {
    return addr >> (offset_bits + index_bits);
}
